/*
Licensed Materials - Property of IBM
5725-C95
(C) Copyright IBM Corporation 2013. All Rights Reserved.
US Government Users Restricted Rights- Use, duplication or disclosure
restricted by GSA ADP Schedule Contract with IBM Corp.
*/

define(["require", "dojo/_base/array", "dojo/_base/connect", "dojo/dom-class", "./logger", "./validationHelper", "./coachViewHelper", "./BPM_CONSTANTS", "./CONSTANTS", "./boundaryEventHelper", "./broadcastHelper", "./commonHelper", "./bindingResolver", "dojo/_base/config"], function(require, array, connect, domClass, logger, validationHelper, coachViewHelper, BPM_CONSTANTS, CONSTANTS, boundaryEventHelper, broadcastHelper, commonHelper, bindingResolver, dojoConfig) {

	var ViewContext = function () {	
		this._init = function(ownerView, system, team, element, binding, option, readonly) {
			// Establish the context object
			this.element = element;
			this.binding = binding;
			this.options = option;
			this.readonly = readonly;
			this.subview = {};
			this.bpm = {"system": system, "team": team};
    		this.viewid = this.element.getAttribute(CONSTANTS.VIEW_ID);
			this._parent = ownerView;
			this._parentView = null;
			this._validationSubscribed = false;
			this.contextRootMap = dojoConfig.App._bpmContextRootMap;
		};
	}		
	
	// do we really need this? investigate later
	ViewContext.prototype.constructor = ViewContext;

	/**
	 * summary:
	 *		escape the text to clean xss attacks
	 *		using .createTextNode to utilize the browser escaping capabilities
	 * txt: the string to escape
	 */
	ViewContext.prototype.htmlEscape = function (txt) {
		var element = document.createElement('span');
		var content = document.createTextNode(txt);
		element.appendChild(content);
		return element.innerHTML;
	};
	
	ViewContext.prototype.parentView = function(){
		return this._parentView;
	};
	
	ViewContext.prototype.triggerCollaboration = function(payload) {
        var element = this.element;
        require(["./controller"], function(controller) {
            controller.customCollaborationHandler(element.getAttribute("id"), payload);
        });
	};
	
	ViewContext.prototype.getInheritedVisibility = function(){
		//find its parent and get parent's visibility
		var visibility = "EDITABLE";
		var parentScope = this.parentView();
		
		if (!parentScope || !parentScope.context || !parentScope.context.options || !parentScope.context.options._metadata)
		      return visibility;
			  
		//if it is null, goes up
		visibility = parentScope.context.options._metadata.visibility.get("value");
		if (visibility == "DEFAULT"){
			return parentScope.context.getInheritedVisibility();
		}
		return visibility;
	};
	
	ViewContext.prototype.bindingType = function() {
		return this.element.getAttribute("data-bindingType");
	};
	
	/*
	Fires a named boundary event. When the boundary event is handled, the supplied callback is called. 
	The callback function is optional.
	callback function has a response object as single parameter - {status: boundaryEventStatus}
	boundaryEventStatus in above response object has one of following 4 values:
		"executed" 		- boundary event has been successfully executed.
		"unqualified"	- context.element doesn't have attribute 'data-eventid', boundary event can not be executed.
		"cancelled"		- boundary event is canceled as the result of calling context.cancelBoundaryEvents() API.
		"abandoned"		- boundary event is abandoned by coach framework when coach framework can not process this boundary event, e.g. failed to communicate with server. 	
	
	options is optional, it contains following properties:
		callbackForAll	- boolean value (true/false) to indicate if the supplied callback should be invoked. The default value is false.
						  If the value is true, callback function (if supplied) is invoked regardless the status is "executed", "unqualified", "cancelled" or "abandoned".
						  If the value is false (or "options" parameter does not have property "callbackForAll", or "options" is not provided), 
						  callback function (if supplied) is invoked only when the status is "executed".
	*/
	ViewContext.prototype.trigger = function(callbackfn, options) {
		var _this = this;
		var alwaysInvokeCallback = false;
		if(!!options && options.callbackForAll === true) {
			alwaysInvokeCallback = true;
		} 
		setTimeout(function() {
			var bdEventId = _this.element.getAttribute('data-eventid');
			if (!!bdEventId) {
				var engine = _this._getEngine();
				var boundaryEventInfo = engine.getBoundaryEventInfo(bdEventId);
				var validateBoundaryEventInfo = engine.getValidateBoundaryEventInfo();
				connect.publish("com.ibm.bpm.coach.BoundaryEventTriggered", [{
					boundaryEventInfo: boundaryEventInfo, validateBoundaryEventInfo: validateBoundaryEventInfo, callback:callbackfn, alwaysInvokeCallback: alwaysInvokeCallback
				 }]);
			} else {
				// 'data-eventid' doesn't exist, so no boundary event should be handled. Anyway, need to let caller know
				boundaryEventHelper.callbackForBoundaryEvent(callbackfn, alwaysInvokeCallback, BPM_CONSTANTS.BOUNDARY_EVENT_STATUS_UNQUALIFIED);
			}
		}, 1);
	};
	
	/*
	 broadcast the supplied message. Input parameter message is an object which is expected to have an attribute "name".
	 This API is not supposed to broadcast sensitive or confidential message.
	*/
	ViewContext.prototype.broadcastMessage = function(message) {
		broadcastHelper.broadcastMessage(message);
	};	
	
	/*
	Return turn if origin matches where Coach is loaded from or one of the sites listed on whitelist, 
	Otherwise false.
	Typical scenario - Coach view receives an postMessage event , coach view can check if the event comes 
	from a trusted site by calling this API with event.origin as parameter.	

	origin:	the site to be checked. Should include protocol, host and port.
			Example value of origin: "https://bpm1.ibm.com:9080", "https://bpm2.ibm.com:9443", "http://bpm3.ibm.com".
	*/
	ViewContext.prototype.isTrustedSite = function(origin) {
		var engine = this._getEngine();
		return engine.checkPostMessageOrigin(origin);
	};	
		
	/*
	Cancel all the boundary events that have not been sent to server.
	Once the boundary event is canceled, the callback function (if supplied) is called to notify the status of boundary event.
	See context.trigger(callback) API for detailed information.
	*/
	ViewContext.prototype.cancelBoundaryEvents = function() {
        require(["./controller"], function(controller) {
            controller.clearBoundaryEventQueue(true);
        });
	};
	
	
	// Public API for view developer.
	/*
	If domNode is the node of any view, a single instance of view will be created and returned.
	If domNode is the node of contextBox, framework creates view instances for all the views owned 
	by this contentBox, and recursively create views for all the framework managed contentBoxs 
	that are owned by this contentBox. An array of view instance will be returned.
	*/
	ViewContext.prototype.createView = function(divNode, index, parentView) {
		logger.log("CoachView", "context.createView", "divNode: " + divNode.id + ", index: " + index + ", parentView: " + parentView);

		/* parentViewScope may not be this._parent, for example,
				viewX
					viewB
						ViewA
							CB_1 (fwk managed)
								ViewC
								CB_2 (view managed, constructed by ViewB)
									ViewD
								CB_3 (fwk managed)
									ViewE
		
		CB_2 is constructed by ViewB, but viewB is not the parentViewScope of the subview (ViewD) in CB_2
		Parent of viewD is View A.

        Parent-child relationship is based on dom structure. ie, your most immediate outer view scope is your parent.
		vice versa: your immediate inner view scope is your child
		
		But at creation time, the view scope thats call viewscope.context.createView  may not be the parent of the view to be created
        This happens to view managed content box		
		
		*/	
		
		var engine = this._getEngine();
		var parentViewScope = (!!(parentView)) ? parentView : coachViewHelper.findParentViewScope(this._parent, divNode);
		
		if(!!index || index == 0) {
			engine.adjustIdAndBinding(divNode, parentViewScope, index);
		}
		
		if(engine.isContentBox(divNode)) {
			var newViewList = new Array();			
			engine.buildContentBox(parentViewScope, divNode, newViewList);
			
			for(var k=0; k<newViewList.length; k++) {
				engine.callView(newViewList[k]);
			}
			
			// check if those newly created view instances have special visibility setting
			if(!!engine.GlobalAssets.coach) {
				for(var i=0; i<newViewList.length; i++) {
					engine.GlobalAssets.coach.updateVisibilityIfApplicable(newViewList[i]);
				}
			}
						
			// check if it's needed to deliver the validation event to the newly created views
			var foundNewValidationEvent = false;
			for(var k=0; k<newViewList.length; k++) {
				foundNewValidationEvent = foundNewValidationEvent || validationHelper.validateView(newViewList[k]);
			}
			if(foundNewValidationEvent) {
				// the new view or its descendant views own validation event
				// just tell engine to deliver the new validation event. It's up to engine to decide when to deliver
				engine.deliverValidationEvents();
			}

			return newViewList;			
		} else {
			// this divNode is expected to be a view
	    	var viewScope = engine.initView.apply(engine.GlobalAssets, [divNode, parentViewScope]);
			engine.callView(viewScope);
			
			// check if this newly created view instance has special visibility setting
			if(!!engine.GlobalAssets.coach) {
				engine.GlobalAssets.coach.updateVisibilityIfApplicable(viewScope);
			}

			// check if it's needed to deliver the validation event to the newly created views
			var foundNewValidationEvent = validationHelper.validateView(viewScope);
			if(foundNewValidationEvent) {
				// the new view or its descendant views own validation event
				// just tell engine to deliver the new validation event. It's up to engine to decide when to deliver
				engine.deliverValidationEvents();
			}
			
			return viewScope;
		}				
	};
	
	// Public API for view developer. 		
	ViewContext.prototype.deleteView = function(divNode) {
		logger.log("CoachView", "context.deleteView", "divNode: " + divNode.id);
		if(!divNode) 	return;
		
		var engine = this._getEngine();
		if(engine.isContentBox(divNode)) {
			engine.deleteContentBox(divNode, this._parent);				
		} else {
			// this divNode is expected to be a view
			engine.deleteView(divNode, this._parent);				
		}
	};
	
	ViewContext.prototype.setDisplay = function(isDisplay, domNode) {
		logger.log("CoachView", "context.setDisplay", "domNode: " + domNode + ", isDisplay: " + isDisplay);
		var engine = this._getEngine();
		var domNode1 = (!!domNode)?domNode:this.element;
    	if (!isDisplay) {
    		domClass.add(domNode1, engine.CONSTANTS.COACH_VIEW_HIDDEN);
    	} else {
    		domClass.remove(domNode1, engine.CONSTANTS.COACH_VIEW_HIDDEN);
    	}
	};
	
	ViewContext.prototype.setVisibility = function(isVisible, domNode) {
		logger.log("CoachView", "context.setVisibility", "domNode: " + domNode + ", isVisible: " + isVisible);
		var engine = this._getEngine();
		var domNode1 = (!!domNode)?domNode:this.element;
    	if (!isVisible) {
    		domClass.add(domNode1, engine.CONSTANTS.COACH_VIEW_INVISIBLE);
    	} else {
    		domClass.remove(domNode1, engine.CONSTANTS.COACH_VIEW_INVISIBLE);
    	}
	};

	ViewContext.prototype.log = function(method, message) {
		var engine = this._getEngine();
		var viewName = engine.CONSTANTS.COACH_VIEW;
		var dataType = this.element.getAttribute(engine.CONSTANTS.VIEW_DATA_TYPE);
		if (!!dataType) {
			dataType = dataType.split(".");
			if (!!dataType) {
				viewName = dataType.pop();
			}
		}
		logger.log(viewName, method, message);
	};		
	
	// sortResult is optional, this API returns unsorted array unless sortResult is set to true
	ViewContext.prototype.getSubview = function(viewId, sortResult) {
		logger.log("CoachView", "context.getSubview", "viewId: " + viewId + ", sortResult: " + sortResult);
		var subviewsArray = new Array();
		var subviewsObject = this.subview[viewId];
		for(var divId in subviewsObject) {
			if(!!subviewsObject[divId]) { // there might be "hole" in subviewsObject as the result of deleteView
				subviewsArray.push(subviewsObject[divId]);
			}
		}
		
		if(sortResult && subviewsArray.length >= 2) {
			var testNode = subviewsArray[0].context.element;
			if(!!testNode.sourceIndex) {
				subviewsArray.sort(function (viewScopeA, viewScopeB) {
					return viewScopeA.context.element.sourceIndex - viewScopeB.context.element.sourceIndex;
				});
			}else if(!!testNode.compareDocumentPosition) {
				subviewsArray.sort(function (viewScopeA, viewScopeB) {
					var nodeA = viewScopeA.context.element;
					var nodeB = viewScopeB.context.element;
					return 3 - (nodeA.compareDocumentPosition(nodeB) & 6);
				});
			}				
		}
		
		return subviewsArray;
	}

	ViewContext.prototype.refreshView = function(){
		logger.log("CoachView", "context.refreshView", "is called");
		this._parent._refreshView(true);
	};
	
	/**
	Calls this API to indicate that it is interested in any validation errors that happen on its descendant views.
	When validation error occurs, validate() method of this view is called prior to that of descendant views.
	
	Views that have data binding are titled to validation event implicitly, namely their validate() method 
	are called when validation event occurs (either "clear" type or "error" type). 
	For the views that don't have data binding, e.g. Tab view, if those views are interested in 
	any validation events that happen on its descendant views, those views have to call this.context.subscribeValidation() 
	to register explicitly.
		
	A view who has data binding can also call this.context.subscribeValidation() to receive the validation events of 
	its descendant views, beside of its own validation event. In this case, after the view calls 
	this.context.unsubscribeValidation(), it can still receive its own validation event.	
	*/
	ViewContext.prototype.subscribeValidation = function(){
		logger.log("CoachView", "context.subscribeValidation", "is called");
		this._validationSubscribed = true;
	};
		
	/**
	Calls this API to unsubscribe.
	Nothing happens if subscribeValidation() was not called previously.
	*/
	ViewContext.prototype.unsubscribeValidation = function(){
		logger.log("CoachView", "context.unsubscribeValidation", "is called");
		this._validationSubscribed = false;
	};
	
	/**
	 * Return true if the view has chance to fall into the troubled situation about different lengths of lists, 
	 * otherwise return false.
	 * For example, table view calls context.containDiffLoopingContext() to determine if table needs to disable its 
	 * add/delete capability during the runtime even if customer has specified to enable add/delete through configuration.
	 * 
	 * Example scenario of different length of lists: Let's say we have following View structure -
	 * 			   TableView (bind to ListA[])
	 * 			       ContentBox
	 * 			           TextView (bind to ListB.currentItem)
	 * 
	 * Case 1: When the length of ListA is shorter than that of ListB, some items in ListB won't have chance to be displayed. 
	 * Case 2: When the length of ListA is longer than that of ListB, some instances of TextView won't have valid data binding. 
	 * Case 3: when he length of ListA and ListB are different, TableView has difficulty to support add/delete. 
	 * 
	 * Above 3 cases are what we mean by "troubled situation about different lengths of lists".
	 * Coach framework handles case 1 and case 2 automatically by showing warning message once the length of involved lists 
	 * are different.
	 * 
	 * TableView can call context.containDiffLoopingContext() to handle case 3. Based on TableView's structure and 
	 * binding information, the return of containDiffLoopingContext() is true. So TableView can disable its 
	 * add and delete capacities accordingly.
	 */
	ViewContext.prototype.containDiffLoopingContext = function(){
		logger.log("CoachView", "context.containDiffLoopingContext", "is called");
		return this._parent._checkIfImpactedByMultipleLists();
	};	
	
	/**
	Get the options object defined in the given viewDomNode even if the view represented by viewDomNode has not been constructed yet. 
	Typically this API is used by the view to have a peek at the configure options (e.g. label, visibility etc) of its children views 
	before any of the children views are created.
	
	viewDomNode - The node of any view.
	
	Following is an example:
		var childOptions = this.context.getOptions(childViewDomNode);
		var childLabel = childOptions._metadata.label.get("value");
		childOptions._metadata.visibility.set("value", "READONLY");
	*/
	ViewContext.prototype.getOptions = function(viewDomNode) {
		logger.log("CoachView", "context.getOptions", "is called for view with DOM ID " + viewDomNode.id);
		var tempViewCfgBinding = viewDomNode.getAttribute(CONSTANTS.DATA_CONFIG);

		var cfgBinding = bindingResolver.globalAssets.options[tempViewCfgBinding]; 
		if (!!cfgBinding) {
			try {
				bindingResolver.resolveOptionsBinding(cfgBinding);
			} catch(ex) {
		    	var errorMsg = "ViewContext.getOptions() caught Exception when calling bindingResolver.resolveOptionsBinding(), Exception: " + ex;
		    	commonHelper.consoleLog(errorMsg, commonHelper.ERROR);
		    	return null;
			}
		}
		return cfgBinding;
	}
	
	// private API
	// get the value of an option
	ViewContext.prototype._getOptionValue = function(configId, optionKey, isMetadata) {
		logger.log("CoachView", "context._getOptionValue", "is called");
		var engine = this._getEngine();
		return engine.getOptionValue(configId, optionKey, isMetadata);
	}

	ViewContext.prototype._getEngine = function(){
		// we know engine module has been loaded.
		return require("./engine");
	};

	/**
	* This API will set data-ibmbpm-uniqueViewId attribute onto the targetNode.  
	* If targetNode is not specified, the attribute will be set to the 
	* this.context.element dom node.
	*/	
	ViewContext.prototype.setUniqueViewId = function(uniqueViewId, targetNode){
		var viewDOM = (!!targetNode)?targetNode:this.element;
		viewDOM.setAttribute("data-ibmbpm-uniqueViewId", uniqueViewId);
	};

	/** 
	* This API will locate and return the first match of the dom node that has the matched data-ibmbpm-uniqueViewId attribute with the input uniqueViewId.  
	* Null will be returned if no match found.  Search will start from startNode, and by default (without the optional optParam.startNode pass 
	* in) startNode is parent DOM node of this.context.element.  Also, search is by default only limited to the scope of the immediate children of the 
	* startNode unless optional optParam.deepSearch = true is passed in. 
	*/	
	ViewContext.prototype.getDomNode = function(uniqueViewId, optParam){
		var startNode = (!!optParam)?optParam['startNode']:null;
		var deepSearch = (!!optParam)?optParam['deepSearch']:null;
		var startingDom = (!!startNode)?startNode:this.element.parentElement;
    	var selector = "* [data-ibmbpm-uniqueViewId='" + uniqueViewId + "']";
    	var targetDOMs = startingDom.querySelectorAll(selector);
    	var targetDOM = null;
		if (!deepSearch) {
			if (!!targetDOMs) {
				for (var i=0; i<targetDOMs.length; i++) {
					if (targetDOMs[i].parentElement == this.element.parentElement) {
						targetDOM = targetDOMs[i];
						break;
					}
				}
			}
		} else {
			if ((!!targetDOMs) && (targetDOMs.length > 0)) {
				targetDOM = targetDOMs[0];
			}
		}
		return targetDOM;
	};
				
	return ViewContext;

});