/*
Licensed Materials - Property of IBM
5725-C95
(C) Copyright IBM Corporation 2012. All Rights Reserved.
US Government Users Restricted Rights- Use, duplication or disclosure
restricted by GSA ADP Schedule Contract with IBM Corp.
*/

// do more test on unbind when we plan to expose this dynamic binding capability as Coach public API
define(["require", "dojo/_base/lang", "dojo/_base/array", "./CONSTANTS", "./commonHelper", "./dataBindingHelper", "./bindingResolver"], function(require, lang, array, CONSTANTS, commonHelper, dataBindingHelper, bindingResolver) {
	
	var dynamicBindingHelper = new Object();
	
	/*	
	 Dynamic bind
	 example of bindingStr: 
	 "local.name", "local.contactBook.owner", 
	 "local.contactBook.phones", "local.contactBook.contacts[index]", "local.contactBook.contacts[index].name", 
	 "local.contactBook.contacts[index].phones[]", "local.contactBook.contacts[index].phones[index].number",
	 "local.contactBook.contacts.listSelected", "local.contactBook.contacts.listSelected.phones.listSelected.name", "local.contactBook.contacts.listSelected.phones[index].name",
	 "config.config1.myOption", "config.config2._metadata.label", "config.config3.myOption.p1",
	 "config.config3.myListOption[index]", "config.config3.myListOption[index].p1",
	 "config.config3.myListOption.listSelected", "config.config3.myListOption.listSelected.p2"
	 
	 "local.contactBook.phones[]" is acceptable, it is interpreted as the recommended notation "local.contactBook.phones". Those 2 notations mean that the change on list itself.
	 "local.contactBook.phones[index]" means the change (add/delete/replace) at List item level.

	  When callback is invoked, input parameter event object is expected to be readonly, should not be modified by callback function.
	 */
	dynamicBindingHelper.bind = function(bindingStr, callback, scope) {
		if(!bindingStr || !callback) {
			return;
		}
			
		if(!!scope) {
			callback = lang.hitch(scope, callback); 
		}
		
		var sourcePathArray = bindingStr.split(".");
		var lastPath = sourcePathArray[sourcePathArray.length - 1];

		// lastPath that ends with "[]" (e.g. "local.contactBook.phones[]") is translated as bare list (e.g. "local.contactBook.phones")
		if(lastPath.indexOf("[]") != -1) {
			sourcePathArray[sourcePathArray.length - 1] = lastPath.replace("[]", "");
			bindingStr = sourcePathArray.join(".");
		}
		
		var handle = {
			"_type": CONSTANTS.HANDLE_TYPE_DYNAMIC,
			"bindingStr": bindingStr,
			"callback": callback, 
			"unbind": function() {
				dynamicBindingHelper.unbindOne(handle);
			}
		};
		
		dynamicBindingHelper._addCallback(bindingStr, callback);

		var bindingPathArray = dynamicBindingHelper._computePaths(bindingStr);
		// bindingStr should always starts with "local"
		dynamicBindingHelper._processBindingPaths(dynamicBindingHelper.rootNode, bindingStr, bindingPathArray, 1);
		
		return handle;		
	}
	
	// perform better than dynamicBindingHelper.unbindOne() in general
	dynamicBindingHelper.unbindMultiple = function(handleArray) {
		if(!!handleArray) {
			var uselessBindingStrArray = [];
			for(var k=0; k<handleArray.length; k++) {
				var uselessBindingStr = dynamicBindingHelper._removeCallback(handleArray[k].bindingStr, handleArray[k].callback);
				if(uselessBindingStr) {
					uselessBindingStrArray.push(uselessBindingStr);
				}
			}
			
			if(uselessBindingStrArray.length > 0) {
				dynamicBindingHelper.rootNode.cleanup(uselessBindingStrArray);
			}
		}
	}
	
	dynamicBindingHelper.unbindOne = function(handle) {
		var uselessBindingStr = dynamicBindingHelper._removeCallback(handle.bindingStr, handle.callback);
		if(uselessBindingStr) {
			dynamicBindingHelper.rootNode.cleanup([uselessBindingStr]);
		}
	}

	// sensitive to the internal tree structure
	// example 1
	// input parameter is: "local.a.b[index].c". output array is: ["root", "local", "a", "b", "$LIST_ELEMENTS$", "b[index]", "c"]
	// example 2
	// input parameter is: 	"config.config1.myOption". output array is: ["root", "config", "config1", "myOption", "value"]
	// example 3
	// input parameter is: 	"config.config2._metadata.label". output array is: ["root", "config", "config2", "_metadata", "label", "value"]
	// example 4
	// input parameter is: 	"config.config3.myOption.p1". output array is: ["root", "config", "config3", "myOption", "value", "p1"]
	// example 5
	// input parameter is: 	"config.config3.myListOption[index].p1". output array is: ["root", "config", "config3", "myListOption", "value", "$LIST_ELEMENTS$", "value[index], "p1"]
	// 					["root", "config", "config3", "myListOption", "$LIST_ELEMENTS$", "myListOption[index], "p1"]
	
	dynamicBindingHelper._computePaths = function(bindingStr) {
		var sourcePathArray = bindingStr.split(".");
		var finalPathArray = [];
		finalPathArray.push(dynamicBindingHelper.VIRTUAL_PATH_ROOT);
		for(var k=0; k<sourcePathArray.length; k++) {
			onePath = sourcePathArray[k];
			
			if(onePath.indexOf(dynamicBindingHelper.INDEX_WITH_BRACKET) != -1) {
				// e.g. a[currentItem]
				// add 2 additional paths to form the internal tree structure
				var virtualPathForList = onePath.replace(dynamicBindingHelper.INDEX_WITH_BRACKET, "");
				finalPathArray.push(virtualPathForList);
				var virtualPathForListContent = dynamicBindingHelper.VIRTUAL_PATH_LIST_CONTENT;				
				finalPathArray.push(virtualPathForListContent);			
			}
			
			finalPathArray.push(onePath);
		}
		
		return finalPathArray;
	}
	
	// provided event is native event received as the result of bind() on Stateful object or bindAll() on List 	
	dynamicBindingHelper._invokeCallback = function(treeNode, event) {
		var callbackList = dynamicBindingHelper.bindingStrToCallbackListMap[treeNode.bindingStr];
		if(!!callbackList) {
			var rtBindingStr = treeNode.computeRtBindingStr();
			if(treeNode.getType() == dynamicBindingHelper.NODE_TYPE_LIST_CONTENT) {
				// change occurs at NODE_TYPE_LIST_CONTENT node, it must be add/delete/set action on list item. need to figure out the index
				if(event.insertedInto != -1) {
					rtBindingStr += "[" + event.insertedInto + "]";
				} else if(event.removedFrom != -1) {
					rtBindingStr += "[" + event.removedFrom + "]";
				}
			}
			for(var k=0; k<callbackList.length; k++) {			
				try {
					var clonedEvent = commonHelper.shallowCloneMap(event);
					clonedEvent.path = rtBindingStr;
					// if this kind of binding capability is exposed as public API in the future, need to monitor the performance of custom script					
					callbackList[k].call(null, clonedEvent);
				} catch(ex) {
					commonHelper.consoleLog(ex, commonHelper.ERROR);
				}
			}
		}
	}
	
	// event objects inside bindingStrToEventGroupMap are artificial events created by dynamicBindingHelper based on the native event
	dynamicBindingHelper._invokeCallback2 = function(bindingStrToEventGroupMap) {
		var bindingStrArray = [];
		for(var bindingStr in bindingStrToEventGroupMap) {
			bindingStrArray.push(bindingStr);
		}
		
		if(bindingStrArray.length == 0) {
			// no event to deliver
			return;
		}
		
		bindingStrArray.sort();
		// depth first
		bindingStrArray.reverse();
		
		for(var i=0; i<bindingStrArray.length; i++) {
			var bindingStr = bindingStrArray[i];
			var callbackList = dynamicBindingHelper.bindingStrToCallbackListMap[bindingStr];
			var rtBindingStrToEventMap = bindingStrToEventGroupMap[bindingStr];
			for(var rtBindingStr in rtBindingStrToEventMap) {
				var event = rtBindingStrToEventMap[rtBindingStr];
				var isBoundObjectSame = dynamicBindingHelper._isBoundObjectSame(event._oldResolvedVar, event._newResolvedVar);
				if(!isBoundObjectSame) {
					event._isResolvedVarChanged = true;
				}
				if(commonHelper.isValuesEqual(event.oldVal, event.newVal) != commonHelper.YES || !isBoundObjectSame) {
					for(var k=0; k<callbackList.length; k++) {			
						try {
							// if this kind of binding capability is exposed as public API in the future, need to monitor the performance of custom script					
							callbackList[k].call(null, event);
						} catch(ex) {
							commonHelper.consoleLog(ex, commonHelper.ERROR);
						}
					}
				}
			}
		}
	}
	
	dynamicBindingHelper._isBoundObjectSame = function(resolvedVar1, resolvedVar2) {
		// if neither resolvedVar1 nor resolvedVar2 exists, treat as they are same
		if(!resolvedVar1 && !resolvedVar2) {
			return true;
		}
		
		if(resolvedVar1 && resolvedVar2) {
			return resolvedVar1.boundObject == resolvedVar2.boundObject;
		}
		return false;
	}

	// sensitive to the internal tree structure
	// example of pathArray: 
	// 			["root", "local", "a", "b", "$LIST_ELEMENTS$", "b[index]", "c"]
	// Note: no need to watch b[index] since its parent node calls bindAll() to take care of its children
	dynamicBindingHelper._processBindingPaths = function(parentTreeNode, bindingStr, pathArray, index) {
		if(index >= pathArray.length) {
			return;
		}

		// _processBindingPaths() is called from many places, e.g. bind(), rebuildAllChildren(), listChanged()
		// we need to make sure that there are solid ground for _processBindingPaths() to execute.
		// For example, when parentTreeNode is NODE_TYPE_LIST_SELECTED and actually no selection yet, _processBindingPaths() should stop asap.
		if(parentTreeNode.getType() == dynamicBindingHelper.NODE_TYPE_LIST_SELECTED) {
			var resolvedVariable =  parentTreeNode.getResolvedBinding();		
			var bindingValue = resolvedVariable.get("value");
			if(!bindingValue) {
				// no selection on the list yet. can't continue. Once "listSelected" is available, we will continue to build this branch
				return;
			}
		}
		
		var onePath = pathArray[index];
			
		if(onePath == dynamicBindingHelper.VIRTUAL_PATH_LIST_CONTENT) {
			// This is the virtual layer that is under list node and holds all list elements
						
			var matchedNode = parentTreeNode.getChildNodeByPath(onePath);
			if(!matchedNode) {
				matchedNode = new TreeNode(parentTreeNode, onePath);
				matchedNode.setType(dynamicBindingHelper.NODE_TYPE_LIST_CONTENT);
			}

			// current path is VIRTUAL_PATH_LIST_CONTENT which is always followed by path "xyz[index]".
			if(index == (pathArray.length - 2)) {
				// VIRTUAL_PATH_LIST_CONTENT node is responsible for the changes on List item level (add/delete/replace)
				matchedNode.watch(bindingStr);
				// we have already handled VIRTUAL_PATH_LIST_CONTENT level, and level "xyz[index]" is the last path. 
				// No need to create node for "xyz[index]" level because VIRTUAL_PATH_LIST_CONTENT is already responsible for the changes at List item level
				return;
			}
			
			// VIRTUAL_PATH_LIST_CONTENT node is responsible for the changes on List item level (add/delete/replace), so need to watch this node
			matchedNode.watch();
			
			if(!matchedNode.firstChild) {
				// need to try to expend the List, because there are other path after "xyz[index]"
				var resolvedBinding = parentTreeNode.getResolvedBinding();
				// resolvedBinding is undefined for following scenario:
				// bindingStr is "local.outerList[index].nameValuePairList[index].name". "local.outerList" contains null as one of its items (let's say index is 1).
				// So the variable instance like "local.outerList[1].nameValuePairList" is not resolvable, as the result, resolvedBinding is undefined.
				if(resolvedBinding) {
					var bindingList = resolvedBinding.get("value");
					// bindingList is undefined for following scenario:
					// bindingStr is "local.outerList[index].nameValuePairList[index].name". "local.outerList" contains empty object {} as one of its items (let's say index is 1). Which is not correct, but it happens.
					// So the variable instance like "local.outerList[1].nameValuePairList" doesn't have value, as the result, bindingList is undefined.
					if(bindingList) {
						var pathOfListItemNode = parentTreeNode.path + dynamicBindingHelper.INDEX_WITH_BRACKET;
						for(var k=0; k<bindingList.length(); k++) {
							new TreeNode(matchedNode, pathOfListItemNode);
						}
					}
				}
			}
			
			var oneNode = matchedNode.firstChild; // oneNode represents list item node
			// matchedNode's child nodes (list items) have been created, and no need to watch on those nodes because 
			// VIRTUAL_PATH_LIST_CONTENT node already takes care of the changes(replace) on list item. so jump over one level. 
			// This is important, otherwise may end up duplicated binding				
			var newIndex = index + 2;
			while(!!oneNode) {
				dynamicBindingHelper._processBindingPaths(oneNode, bindingStr, pathArray, newIndex);
				oneNode = oneNode.next;
			}
		} else {
			var matchedNode = parentTreeNode.getChildNodeByPath(onePath);
			if(!matchedNode) {
				matchedNode = new TreeNode(parentTreeNode, onePath);
			}
			
			if(onePath == "listSelected") {
				// make sure it is "listSelected" of a real list.
				var parentResolvedVariable =  parentTreeNode.getResolvedBinding();		
				var parentBindingValue = parentResolvedVariable.get("value");
				if(parentBindingValue instanceof com.ibm.bpm.coach.List) {
					// yes, "listSelected" of a list
					matchedNode.setType(dynamicBindingHelper.NODE_TYPE_LIST_SELECTED);
										
					var resolvedVariable = parentTreeNode.getResolvedBinding();
					// cache the plainValue of dynamicBindingHelper.NODE_TYPE_LIST_SELECTED node
					matchedNode.plainValue = resolvedVariable.get("value")[matchedNode.path];
				}
			}
			
			if(index == (pathArray.length - 1)) {
				// Last path
				matchedNode.watch(bindingStr);
				return;
			}

			// we still need to watch this node in order to automatically rebind the children nodes when this node is changed
			// scenario: bidingStr is a.b.c, so c is watched explicitly.  Need to watch a and b internally, as the result, when a or b is changed, we have chance to rebind c
			// don't watch following nodes: root, root.local, root.config, root.config.configXYZ,  root.config.configXYZ._metadata
			if(index > 1) {
				// already filtered out root, root.local, root.config
				if(pathArray[1] != "config") {
					matchedNode.watch();
				} else {
					if(index > 2 && onePath != "_metadata") {
						matchedNode.watch();
					}
				}
			}			
			
			dynamicBindingHelper._processBindingPaths(matchedNode, bindingStr, pathArray, index+1);
		}			
	}
	
	dynamicBindingHelper._addCallback = function(bindingStr, callback) {
		var callbackList = dynamicBindingHelper.bindingStrToCallbackListMap[bindingStr];
		if(!callbackList) {
			callbackList = [];
			dynamicBindingHelper.bindingStrToCallbackListMap[bindingStr] = callbackList;
		}
		callbackList.push(callback);
	}
	
	// returns specified bindingStr if that bindingStr doesn't have any callback
	dynamicBindingHelper._removeCallback = function(bindingStr, callback) {
		var isBindingStrUseless = false;
		var callbackArray = dynamicBindingHelper.bindingStrToCallbackListMap[bindingStr];
		if(!!callbackArray) {
			var index = array.indexOf(callbackArray, callback);
			if(index != -1) {
				callbackArray.splice(index, 1);
				if(callbackArray.length == 0) {
					isBindingStrUseless = true;
					dynamicBindingHelper.bindingStrToCallbackListMap[bindingStr] = null;
					delete dynamicBindingHelper.bindingStrToCallbackListMap[bindingStr];
				}
			}
		}
		
		if(isBindingStrUseless) {
			return bindingStr;
		} else {
			return null;
		}
	}


	/*
		treeNode structure:
		{
			path:			"abc",
			bindingStr:		"local.xyz.abc",
			resolvedVar:	resolvedVariable,
			nativeBindingHandle: bindingHandle,
			parent:			parentTreeNode,
			next:			nextSibling,
			previous:		previousSibling,
			firstChild:		firstChildTreeNode,
			lastChild:		lastChildTreeNode,
		}
	*/
	// index is optional
	var TreeNode = function(parentTreeNode, path, index) {
		this.parent = parentTreeNode;
		this.path = path;
		
		if(parentTreeNode) {			
			if(!parentTreeNode.firstChild) {
				// this has to be the first child, even if index is specified with a non-zero value
				parentTreeNode.firstChild = this;
				parentTreeNode.lastChild = this;
				this.previous = null;
				this.next = null;
			} else {
				if(index == 0) {
					var myNext = parentTreeNode.firstChild;
					parentTreeNode.firstChild = this;
					this.previous = null;
					this.next = myNext;
				} else if(!!index){
					// try to put the new node at the desired index
					var nodeAtIndex = parentTreeNode.firstChild;
					for(var k=0; k<index; k++) {
						nodeAtIndex = nodeAtIndex.next;
						if(!nodeAtIndex) {
							break;
						}
					}
					
					if(!!nodeAtIndex) {
						var previousNode = nodeAtIndex.previous;
						previousNode.next = this;
						this.next = nodeAtIndex;
						nodeAtIndex.previous = this;
					} else {
						// specified index is out of bound. add this as the last child of parent node
						parentTreeNode.lastChild.next = this;
						this.previous = parentTreeNode.lastChild;
						parentTreeNode.lastChild = this;
						this.next = null;						
					}
					
				} else {
					// index is not specified, add this as the last child of parent node
					parentTreeNode.lastChild.next = this;
					this.previous = parentTreeNode.lastChild;
					parentTreeNode.lastChild = this;
					this.next = null;
				}
			}
		}		
	}
	
	// dispose this object
	TreeNode.prototype.dispose = function() {
		this.parent = null;
		this.previous = null;
		this.next = null;
		this.firstChild = null;
		this.lastChild = null;

		this.unwatch();
		this.resolvedVar = null;
		this.path = null;
		this.bindingStr = null;
	}
	
	// get resolved binding.
	TreeNode.prototype.getResolvedBinding = function() {
		if(!this.resolvedVar) {
			var rtBindingStr = this.computeRtBindingStr();
			this.resolvedVar =  bindingResolver.resolveBinding(rtBindingStr);
		}
		return this.resolvedVar;
	}
	
	// add native binding listener
	// if bindingStr is specified, means that the node is watched for the user of dynamicBindingHelper.bind()
	// if bindingStr is not specified, means that the node is watched for internal purpose, 
	// e.g. rebind (a.b.c, a is replaced, need to know that change in order to rebind on c), bind (e.g. new item is added to list)
	TreeNode.prototype.watch = function(bindingStr) {
		if(!this.bindingStr && !!bindingStr) {
			this.bindingStr = bindingStr;
		}
		
		if(!this.nativeBindingHandle) {
			var resolvedVariable = this.getResolvedBinding();		
			// bind on it if exists
			if(!!resolvedVariable) {
				if(this.getType() == dynamicBindingHelper.NODE_TYPE_LIST_CONTENT) {
					// only NODE_TYPE_LIST_CONTENT can use bindAll to minitor the change at list item level
					var bindingValue = resolvedVariable.get("value");
					if(bindingValue instanceof com.ibm.bpm.coach.List) {
						this.nativeBindingHandle = bindingValue.bindAll(this.listChange, this);
					}				
				} else {
					if(!!resolvedVariable.property) {
						this.nativeBindingHandle = resolvedVariable.boundObject.bind(resolvedVariable.property, this.propertyChange, this);
					}
				}
			}		
		}
	}
	
	// remove native binding listener
	TreeNode.prototype.unwatch = function() {
		this.bindingStr = null;
		
		if(!!this.nativeBindingHandle) {
			try {
				this.nativeBindingHandle.unbind();
			} catch(ex) {
		    	commonHelper.consoleLog("dynamicBindingHelper.TreeNode.unwatch(), failed to unbind the binding handler, Exception: " + ex, commonHelper.ERROR);
			}
		}
		this.nativeBindingHandle = null;
	}
	
	// sensitive to the internal tree structure
	TreeNode.prototype.computeRtBindingStr = function(index) {
		var pathArray = [];
		var oneNode = this;
		var onePath;
		
		if(!oneNode.parent) {
			// this is root node which has no binding
			return null;
		}
		
		do {
			if(oneNode.getType() == dynamicBindingHelper.NODE_TYPE_LIST_CONTENT) {
				// I am artificial node that groups the list items
				pathArray.push(oneNode.parent.path);
				// skip one level above: list node
				oneNode = oneNode.parent;
			} else if(!!oneNode.path) {			
				if(oneNode.path.indexOf(dynamicBindingHelper.INDEX_WITH_BRACKET) != -1) {
					// e.g. a[index]
					// figure out my position in the list
					var index = oneNode.getIndex();
					var rtPath = oneNode.path.replace(dynamicBindingHelper.INDEX_WITH_BRACKET, "[" + index + "]");
					pathArray.push(rtPath);
					// skip 2 levels above: VIRTUAL_PATH_LIST_CONTENT node and List node
					oneNode = oneNode.parent.parent;	
				} else {
					pathArray.push(oneNode.path);
				}
			}
			oneNode = oneNode.parent;
		} while(!!oneNode.parent);
		
		return pathArray.reverse().join(".");
	}
	
	// sensitive to the internal tree structure
	TreeNode.prototype.computeBindingStr = function() {
		var pathArray = [];
		var oneNode = this;
		var onePath;
		
		if(!oneNode.parent) {
			// this is root node which has no binding
			return null;
		}
		
		do {
			if(oneNode.getType() == dynamicBindingHelper.NODE_TYPE_LIST_CONTENT) {
				// I am artificial node that groups the list items
				pathArray.push(oneNode.parent.path + dynamicBindingHelper.INDEX_WITH_BRACKET);
				// skip one level above: list node
				oneNode = oneNode.parent;
			} else if(!!oneNode.path) {			
				pathArray.push(oneNode.path);
				if(oneNode.path.indexOf(dynamicBindingHelper.INDEX_WITH_BRACKET) != -1) {
					// e.g. a[index]
					// skip 2 levels above: VIRTUAL_PATH_LIST_CONTENT node and List node
					oneNode = oneNode.parent.parent;	
				}
			}
			oneNode = oneNode.parent;
		} while(!!oneNode.parent);
		
		return pathArray.reverse().join(".");
	}
	
	
	// return the position of this node among its siblings
	TreeNode.prototype.getIndex = function() {
		var index = 0;
		var tempNode = this;
		while(!!tempNode.previous) {
			index++;
			tempNode = tempNode.previous;
		}

		return index;
	}

	TreeNode.prototype.getChildNodeByIndex = function(index) {
		var childNode = this.firstChild;
		for(var k=0; k<index; k++) {
			if(!!childNode) {
				childNode = childNode.next;
			} else {
				return null;
			}
		}
		
		return childNode;
	}

	// get the first child node that matches specified path
	TreeNode.prototype.getChildNodeByPath = function(path) {
		var oneNode = this.firstChild;
		while(!!oneNode) {
			if(oneNode.path == path) {
				return oneNode;
			}
			oneNode = oneNode.next;
		}
		return null;
	}
	
	// remove one child node
	TreeNode.prototype.removeChild = function(childNode) {
		//
		childNode.removeAllChildren();
		
		var leftSide = childNode.previous;
		var rightSide = childNode.next;
		
		if(!!leftSide) {
			leftSide.next = rightSide;
		} else {
			// childNode is the first child
			this.firstChild = rightSide;
		}
		
		if(!!rightSide) {
			rightSide.previous = leftSide;
		} else {
			// childNode is the last child
			this.lastChild = leftSide;
		}
		
		// now, no node links to childNode
		childNode.dispose();
	}

	// remove all children nodes
	TreeNode.prototype.removeAllChildren = function() {
		var childToRemove = this.firstChild;
		
		while(!!childToRemove) {
			childToRemove.removeAllChildren();
			var nextOne = childToRemove.next;
			childToRemove.dispose();
						
			childToRemove = nextOne;
		}
		
		this.firstChild = null;
		this.lastChild = null;
	}
	
	// Since some bindingStr are not used any more (as the result of unbind action), clean up the branch start from myself
	TreeNode.prototype.cleanup = function(uselessBindingStrArray) {
		// firstly check if my branch needs to take any action
		myPath = this.path;
		var bindingStrToCheck = this.bindingStr ? this.bindingStr : this.computeBindingStr();
		
		var worthToContinue = false;
		if(this == dynamicBindingHelper.rootNode) {
			// I am root node which is just an artificial node and doesn't match to any binding path. So need to go head to check my child nodes.
			worthToContinue = true;
		} else {
			for(var k=0; k<uselessBindingStrArray.length; k++) {
				if(commonHelper.startsWithPath(uselessBindingStrArray[k], bindingStrToCheck, true)) {
					worthToContinue = true;
					break;
				}
			}
		}
		
		if(!worthToContinue) {
			return;
		}
		
		// cleanup children node 
		var childNode = this.firstChild;		
		while(!!childNode) {
			var temp = childNode.next;
			childNode.cleanup(uselessBindingStrArray);
			childNode = temp;
		}
		
		// cleanup myself
		
		// // check if need to remove this.bindingStr. the node may still be watched after removing this.bindingStr. for example, in order to support rebind
		if(this.bindingStr && array.indexOf(uselessBindingStrArray, this.bindingStr) != -1) {
			this.bindingStr = null;
		}
		
		if(this.firstChild) {
			// I still have child node, so I have reason to stay alive
			return;
		}
		
		if(this.bindingStr) {
			// I still have this.bindingStr, means that some real callback relies on me. So stay alive
			return;
		}
				
		if(this.getType == dynamicBindingHelper.NODE_TYPE_LIST_CONTENT) {
			// I am artificial node that groups List items.
			// I don't have child, don't have this.bindingStr, but I should be alive as long as somebody still need me to watch the change (add/delete/replace) on list item level.
			for(var subscribedBindingStr in dynamicBindingHelper.bindingStrToCallbackListMap) {
				if(commonHelper.startsWithPath(subscribedBindingStr, bindingStrToCheck, true)) {
					return;
				}
			}
		}
		
		// I have no value any more, ask parent to remove me
		if(this.parent) {
			this.parent.removeChild(this); // removeChild() calls unwatch() implicitly
		}
	}
	
	TreeNode.prototype.setType = function(type) {
		this.type = type;
	}
	TreeNode.prototype.getType = function() {
		return this.type;
	}
	
	TreeNode.prototype.getLevel = function() {
		var myLevel = 0;
		var temp = this;
		while(!!temp.parent) {
			myLevel++;
			temp = temp.parent;
		}
		return myLevel;
	}
	
	// sensitive to the internal tree structure
	/*
	 skipNode: if a node is skipped, plainValue is passed down to its children nodes directly.
	           node is skipped for 2 scenarios - a) starting node, because it has native event already. b) dynamicBindingHelper.VIRTUAL_PATH_LIST_CONTENT node because nobody could bind on it
	 plainValue: is used to find the value for event.oldVal or event.newVal 
	 bindingStrToEventGroupMap: a map between bindingStr and an object which is the map between rtBindingStr and changeEvent.
	 isForOldVar: if it is true, collect old value for the change event. Otherwise, collect new value for the change event.
	*/
	TreeNode.prototype.collectInfoForChangeEvents = function(skipNode, plainValue, bindingStrToEventGroupMap, isForOldVar, indexIfListItem) {
		// handle this node firstly
		if(!skipNode) {
			if(this.bindingStr && this.path != dynamicBindingHelper.VIRTUAL_PATH_LIST_CONTENT) {
				// work out a property change event
				var rtBindingStr = this.computeRtBindingStr();
				if(!bindingStrToEventGroupMap[this.bindingStr]) {
					bindingStrToEventGroupMap[this.bindingStr] = {};
				}
				if(!bindingStrToEventGroupMap[this.bindingStr][rtBindingStr]) {
					bindingStrToEventGroupMap[this.bindingStr][rtBindingStr] = {path: rtBindingStr, oldVal: null, newVal: null, property: this.path, _origin: "system", _oldResolvedVar: null, _newResolvedVar: null};
				}
				
				if(isForOldVar) {
					bindingStrToEventGroupMap[this.bindingStr][rtBindingStr].oldVal = plainValue;
					bindingStrToEventGroupMap[this.bindingStr][rtBindingStr]._oldResolvedVar = this.resolvedVar;
				} else {
					bindingStrToEventGroupMap[this.bindingStr][rtBindingStr].newVal = plainValue;
					bindingStrToEventGroupMap[this.bindingStr][rtBindingStr]._newResolvedVar = this.resolvedVar;
				}
			} else if((this.path.indexOf(dynamicBindingHelper.INDEX_WITH_BRACKET) != -1) && this.parent.bindingStr) {
				// work out a list item change event (add/delete/replace list item)
				var bindingStr = this.parent.bindingStr;
				var rtBindingStr = this.computeRtBindingStr();
				if(!bindingStrToEventGroupMap[bindingStr]) {
					bindingStrToEventGroupMap[bindingStr] = {};
				}
				if(!bindingStrToEventGroupMap[bindingStr][rtBindingStr]) {
					bindingStrToEventGroupMap[bindingStr][rtBindingStr] = {path: rtBindingStr, oldVal: null, newVal: null, insertedInto: -1, removedFrom: -1, _origin: "system"};
				}
				
				if(isForOldVar) {
					bindingStrToEventGroupMap[bindingStr][rtBindingStr].oldVal = plainValue;
					bindingStrToEventGroupMap[bindingStr][rtBindingStr].removedFrom = indexIfListItem;
					bindingStrToEventGroupMap[bindingStr][rtBindingStr]._oldResolvedVar = this.resolvedVar;
				} else {
					bindingStrToEventGroupMap[bindingStr][rtBindingStr].newVal = plainValue;
					bindingStrToEventGroupMap[bindingStr][rtBindingStr].insertedInto = indexIfListItem;
					bindingStrToEventGroupMap[bindingStr][rtBindingStr]._newResolvedVar = this.resolvedVar;
				}
			}
		}
		
		// handle the children nodes
		var child = this.firstChild;
		
		while(!!child) {	
			var isListContentNode = (child.path == dynamicBindingHelper.VIRTUAL_PATH_LIST_CONTENT); 
			var childPlainValue = null;
			var childIndex = null;
			if(!!plainValue) {
				if(isListContentNode) {
					childPlainValue = plainValue.items;
				} else if(child.path.indexOf(dynamicBindingHelper.INDEX_WITH_BRACKET) != -1) {
					// e.g. a[index]
					// figure out my position in the list
					childIndex = child.getIndex();
					// plainValue is expected to be an Array
					childPlainValue = plainValue[childIndex];
				} else {
					// property of complex object
					childPlainValue = plainValue[child.path];
				}
			}
			
			if(child.getType() == dynamicBindingHelper.NODE_TYPE_LIST_SELECTED && !childPlainValue) {
				// native change event doesn't contain the value for listSelected. To work around it, dynamicBindingHelper.NODE_TYPE_LIST_SELECTED tree node caches the plain value by itself
				childPlainValue	= child.plainValue;
			}
			
			child.collectInfoForChangeEvents(isListContentNode, childPlainValue, bindingStrToEventGroupMap, isForOldVar, childIndex);
			child = child.next;
		}
	}
	
	TreeNode.prototype.rebuildAllChidren = function() {
		this.removeAllChildren();

		var bindingStr = !!this.bindingStr ? this.bindingStr : this.computeBindingStr();
		
		var affectedBindingStrArray = [];
		for(var subscribedBindingStr in dynamicBindingHelper.bindingStrToCallbackListMap) {
			if(commonHelper.startsWithPath(subscribedBindingStr, bindingStr, true)) {
				affectedBindingStrArray.push(subscribedBindingStr);
			}
		}
		if(affectedBindingStrArray.length > 0) {
			var nextLevel = this.getLevel() + 1;
			for(var k=0; k<affectedBindingStrArray.length; k++) {
				var bindingPathArray = dynamicBindingHelper._computePaths(affectedBindingStrArray[k]);
				dynamicBindingHelper._processBindingPaths(this, affectedBindingStrArray[k], bindingPathArray, nextLevel);
			}
		}
	}
	
	// sensitive to the internal tree structure
	TreeNode.prototype.propertyChange = function(event) {
		if(event.property == "value" && event.property != this.path) {
			// sometimes (e.g. binding is on config) event.property is "value" although what has been changed really has a meaningful name.
			event.property = this.path;
		}
		
		// before removing the children node, walk though entire branch to collect the information (mainly event.oldVal) for the possible change event
		var bindingStrToEventGroupMap = {};
		this.collectInfoForChangeEvents(true, event.oldVal, bindingStrToEventGroupMap, true);
	
		// this node is changed. If this node has chidren nodes, the native listener on the children nodes need to be refreshed.
		this.rebuildAllChidren();
					
		// now the new branch is well established. walk though entire branch to collect the information (mainly event.NewVar) for the possible change event
		this.collectInfoForChangeEvents(true, event.newVal, bindingStrToEventGroupMap, false);
		
		// fire the change events per this structure change
		dynamicBindingHelper._invokeCallback2(bindingStrToEventGroupMap);

		if(this.bindingStr) {
			dynamicBindingHelper._invokeCallback(this, event);
		}
		
		if(this.getType() == dynamicBindingHelper.NODE_TYPE_LIST_SELECTED) {
			// cache the plainValue for dynamicBindingHelper.NODE_TYPE_LIST_SELECTED node
			this.plainValue = event.newVal;
		}
	}
	
	// sensitive to the internal tree structure
	// listener for List change, "this" means a VIRTUAL_PATH_LIST_CONTENT node
	TreeNode.prototype.listChange = function(changeEvent) {
		// bindingStrToEventGroupMap is responsible for storing event information
		var bindingStrToEventGroupMap = {};
		if(changeEvent.removedFrom >= 0 && changeEvent.insertedInto >= 0 && changeEvent.removedFrom == changeEvent.insertedInto) {
			// set action on List item
			var nodeToUpdate = this.getChildNodeByIndex(changeEvent.removedFrom);
			
			if(nodeToUpdate) {
				// before removing the children, walk though entire branch to collect the information (mainly event.oldVal) for the possible change event
				nodeToUpdate.collectInfoForChangeEvents(true, changeEvent.oldVal, bindingStrToEventGroupMap, true);
	
				// nodeToUpdate is changed. If it has chidren nodes, the native listener on the children nodes need to be refreshed.
				nodeToUpdate.rebuildAllChidren();
				
				// now the new branch is well established. walk though entire branch to collect the information (mainly event.NewVar) for the possible change event
				nodeToUpdate.collectInfoForChangeEvents(true, changeEvent.newVal, bindingStrToEventGroupMap, false);
			}
		} else {
			if(changeEvent.removedFrom >= 0) { 
				// one item has been removed from the list.								
				var nodeToRemove = this.getChildNodeByIndex(changeEvent.removedFrom);
				
				if(nodeToRemove) {
					// before removing the child, walk though entire branch headed by this child node to collect the information for the possible change event
					nodeToRemove.collectInfoForChangeEvents(true, changeEvent.oldVal, bindingStrToEventGroupMap, true);
	
					this.removeChild(nodeToRemove);
				}
			}
			
			if(changeEvent.insertedInto >= 0) {
				// one item has been added to the list.	
				// need to check if the future change on some node of this new branch needs to be watched
				
				var listNode = this.parent;
				// the path for the newly added list item
				var path = listNode.path + dynamicBindingHelper.INDEX_WITH_BRACKET;
				var grandparentNodeBindingStr = !!listNode.parent.bindingStr ? listNode.parent.bindingStr : listNode.parent.computeBindingStr();
				// the prefix of bindingStr for the newly added list item
				var prefix = grandparentNodeBindingStr + "." + path;
				
				var affectedBindingStrArray = [];
				for(var subscribedBindingStr in dynamicBindingHelper.bindingStrToCallbackListMap) {
					// prefix is something like local.a.b[index] which is already monitored by VIRTUAL_PATH_LIST_CONTENT node.
					// so, if subscribedBindingStr is same as prefix, don't need to handle it since "this" (VIRTUAL_PATH_LIST_CONTENT node) is responsible for this kind of binding
					if(commonHelper.startsWithPath(subscribedBindingStr, prefix)) {
						affectedBindingStrArray.push(subscribedBindingStr);
					}
				}
				
				var newNode = null;
				if(affectedBindingStrArray.length > 0) {
					// need to handle the new branch
					// the parent node for the new node is myself, VIRTUAL_PATH_LIST_CONTENT node
					var parantForNewNode = this;
					newNode = new TreeNode(parantForNewNode, path, changeEvent.insertedInto);
					var nextLevel = newNode.getLevel() + 1;
					for(var k=0; k<affectedBindingStrArray.length; k++) {
						var bindingPathArray = dynamicBindingHelper._computePaths(affectedBindingStrArray[k]);
						dynamicBindingHelper._processBindingPaths(newNode, affectedBindingStrArray[k], bindingPathArray, nextLevel);
					}
				}
				if(newNode) {
					// now the new branch is well established. walk though entire branch to collect the information (mainly event.NewVar) for the possible change event
					newNode.collectInfoForChangeEvents(true, changeEvent.newVal, bindingStrToEventGroupMap, false);
				}			
			}
		}

		// fire the change events per this structure change
		dynamicBindingHelper._invokeCallback2(bindingStrToEventGroupMap);

		if(this.bindingStr) {
			// if this.bindingStr exists, means there is at least one callback
			dynamicBindingHelper._invokeCallback(this, changeEvent);
		}				
	}
		
	dynamicBindingHelper.NODE_TYPE_LIST_SELECTED = "TYPE_LIST_SELECTED";
	dynamicBindingHelper.NODE_TYPE_LIST_CONTENT = "TYPE_LIST_CONTENT";

	dynamicBindingHelper.INDEX_WITH_BRACKET = CONSTANTS.INDEX_WITH_BRACKET;
	dynamicBindingHelper.VIRTUAL_PATH_LIST_CONTENT = "$LIST_ELEMENTS$";
	dynamicBindingHelper.VIRTUAL_PATH_ROOT = "$ROOT$";
	
	dynamicBindingHelper.rootNode = new TreeNode(null, dynamicBindingHelper.VIRTUAL_PATH_ROOT);
	
	dynamicBindingHelper.bindingStrToCallbackListMap = {};
	
	return dynamicBindingHelper;
});
